import plotly.express as px
import plotly.graph_objects as go
import pandas as pd
from umap import UMAP
from sklearn.manifold import TSNE
"""
    Copyright (C) 2023  Ramin Mohammadi

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses>."""


def Visualize_Distance_GeneExp(df: pd.DataFrame, cell_types: list, color_scale: dict, umap: True, n_neighbors=10, min_dist=0.5):
    """
    UMAP or TSNE plot each cell by condensing the euclidean distances to each gene (allowing each point in the plot to represent a cell). Then, every cell/point 
    is labelled with either the predicted cell type or actual cell type.
    # https://pypi.org/project/umap-learn/
    # umap-learn has requirement packages, make sure to pip install those as well (look at above link)
        input:
            - df: pandas data frame of euclidean distances between genes and cells where rows are cells and genes are columns
            - cell_types: list of the labelled cell types for each cell whether it be the actual or predicted labels
            - color_scale: dictionary where key is the cell type/gene set name and the value is the color that will be used for that label (color names are from plotly)
            - umap: if True, plots using UMAP. Otherwise, TSNE
            - n_neighbors: (only applies for UMAP) should be a value between 2 and 100. Low values will push UMAP to focus more on local structure by constraining the number 
            of neighboring points considered when analyzing the data in high dimensions, while high values will push UMAP towards representing the big-picture structure 
            while losing fine detail.
            - min_dist: (only applies for UMAP) should be a value between 0.0 and 0.99. Controls how tightly UMAP clumps points together, with low values leading to more 
            tightly packed embeddings. Larger values of min_dist will make UMAP pack points together more loosely, focusing instead on the preservation of the broad topological structure
        output:
            - plot opened in a web browser
    """    
    
    if cell_types is None:
        print('ERROR: list of cell types for individual cells is not given')
        return
    if color_scale is None:
        print('ERROR: dictionary of color map for coloring of cell types is not given')
        return  
    if umap is True: # perform UMAP
        reducer = UMAP(n_neighbors=n_neighbors, min_dist=min_dist, n_components=2, init='random', random_state=0)
        embedding = reducer.fit_transform(X=df)
    else: # perform TSNE
        embedding = TSNE(n_components=2, learning_rate='auto', init='random', perplexity=3).fit_transform(X=df)
        
    # add cell type as a column so can color code the plot by cell types
    plot_df = pd.DataFrame(embedding, columns=["x", "y"])
    plot_df["cell_type"] = cell_types
    
    # scatter plot the UMAP using plotly
    fig = px.scatter(plot_df, x="x", y="y", color="cell_type", color_discrete_map=color_scale)
    fig.show()


def Visualize_Coordinates(cellCoordinates: pd.DataFrame, geneCoordinates: pd.DataFrame, n_neighbors=15, min_dist=0.1):
    """
    Visualize j dimensions of cell and gene coordinates in a 3D space (j dimensions condensed to 3). Utilizes plotly for plotting and 
    umap-learn's UMAP function to condense the dimensions.
        input:
            - cellCoordinates: pandas data frame with cells as rows and dimensions as columns
            - geneCoordinates: pandas data frame with genes as rows and dimensions as columns
            - n_neighbors: should be a value between 2 and 100. Low values will push UMAP to focus more on local structure by constraining the number of neighboring 
            points considered when analyzing the data in high dimensions, while high values will push UMAP towards representing the big-picture structure while losing fine detail.
            - min_dist: should be a value between 0.0 and 0.99. Controls how tightly UMAP clumps points together, with low values leading to more tightly packed embeddings. 
            Larger values of min_dist will make UMAP pack points together more loosely, focusing instead on the preservation of the broad topological structure.
        output:
            - opens a link to the umap plot in a web browser. .html file of the plot is saved locally
    """
    
    print("\nCreating UMAP of cell and gene coordinates in 3 dimensional space...\n")
    # construct umap object with parameters you want to be performed (n_components is number of dimensions you want the data to be reduced to) 
    umap_3d = UMAP(n_neighbors=n_neighbors, min_dist=min_dist, n_components=3, init='random', random_state=0)
    
    # Train our reducer, letting it learn about the manifold.
    # Fit X into an embedded space and return that transformed output.
    # the proj variables are n rows by 3 columns NDarray (each column is a dimension in 3D condensed representation from UMAP)
    # the values in the outputted ndarray change b/c the result is a representation of clusters/neighboring points meaning
    # the values outputted from umap are different than the magnitude of the orginal data b/c values used to graph visual clusters to show denisty of data
    
    # perform specifiied UMAP on data
    proj_3d_gene = umap_3d.fit_transform(X=geneCoordinates) 
    proj_3d_cell = umap_3d.fit_transform(X=cellCoordinates)
    
    # plot the results column by column (first column is x, 2nd column is y, and if have 3 dimensions, 3rd column is z)
    # each row in proj_3d_gene corresponds to a gene (gene list based on how was ordered in original data before performing umap),
    # likewise for proj_3d_cell where every row in it corresponds to a cell
    # THEREFORE, every point in the 3D graph is either an individual gene or cell
    fig = go.Figure(data=[
        go.Scatter3d(
            x=proj_3d_gene[:, 0], y=proj_3d_gene[:, 1], z=proj_3d_gene[:, 2],
            mode='markers',
            text=geneCoordinates.index.to_list(),  # text attribute labels every row where every row is an x,y,z coordinate (label every coordinate with gene name)
            name="Gene Coordinate",
            marker = dict(
                size=6,
                opacity=.5,
                colorscale=px.colors.sequential.Bluyl)
        ),
        go.Scatter3d(
            x=proj_3d_cell[:, 0], y=proj_3d_cell[:, 1], z=proj_3d_cell[:, 2],
            mode='markers', 
            text=cellCoordinates.index.to_list(), 
            name="Cell Coordinate",
            marker = dict(
                size=6,
                opacity=.5,
                colorscale=px.colors.sequential.Hot)
        )     
    ])
    
    fig.update_layout(scene = dict(
                    xaxis_title='MCA_1',
                    yaxis_title='MCA_2',
                    zaxis_title='MCA_3'),
                    )
    
    fig.show()
    fig.write_html("coordinate.html", include_plotlyjs=True, full_html=True)


import numpy as np

"""
    Copyright (C) 2023  Ramin Mohammadi

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses>."""
    
# Reference: https://github.com/RausellLab/CelliD


def MCAStep1(arr: np.ndarray) -> dict:    
    """
    Computes fuzzy matrix and matrix of standardized relative frequencies
    input: 
        - arr: a 2D numpy array with cells as rows and genes as columns, consisting of preprocessed raw gene count data
    returns a dictionary:
        - Z: S_N,K, matrix of standardized relative frequencies
        - D_r: D_r ^-1/2, vector representing diagonal matrix of row sums of the matrix of relative frequencies ^-1/2
        - D_c: D_c ^-1/2, vector representing diagonal matrix of column sums of the matrix of relative frequencies ^-1/2
        - X: fuzzy-coded indicator matrix 
    """
    
    ############## Generate fuzzy-coded indicator matrix: ############
    # rmin , rmax , and range_list are vectors
    rmin = np.min(arr, axis=0) # get min val in each gene column -> min(M_p) 
    rmax = np.max(arr, axis=0) # get max val in each gene column -> max(M_p)
    range_list = rmax - rmin  # max(M_p) - min(M_p) for each gene row p
    # m_np − min(M_p) -> subtract the min value for a row from each scalar in a row:
    for index in range(len(rmin)): # iterate column by column and value by value in rmin
        arr[:, index] -= rmin[index]
    # m_np − min(M_p) / max(M_p) - min(M_p)
    for index in range(len(range_list)):
        arr[:, index] /= range_list[index]
        arr[:, index] = np.nan_to_num(arr[:, index]) # turns nan values to 0 in case where a number is divided by 0
    
    # Fuzzy matrix stored in FM, size: N cells, K = 2P genes 
    # below puts fuzzy matrix in cellid paper intended format of complementary values (by columns)
    complement = 1 - arr
    rows, columns = np.shape(arr)
    FM = np.empty(shape=(rows, columns * 2)) # fuzzy matrix ends up with total of K = 2P categories/columns
    index = 0
    for column in range(0, np.shape(FM)[1]):
        if column % 2 == 0:
            FM[:, column] = arr[:, index] # x_p+ stored in even columns
        else:
            FM[:, column] = complement[:, index] # x_p- complement stored in odd columns
            index += 1
    X=FM # fuzzy matrix stored for later reference
    # -------------------------------------- End of generating fuzzy matrix 
    
    ############## Generate matrix of standardized relative frequencies #########
    # S = D_r ^−1/2 * FM * D_c ^−1/2  
    # OR  
    # S = (1 / D_r ^1/2) * FM  * (1 / D_c ^1/2) -> the inverse of diagonal matrices are simply the reciprocal of each value
       
    # Matrix of relative frequencies: 
    # F_N,K = 1/NP * X , where there are N cells and P genes
    FM = np.divide(FM , (np.shape(arr)[0] * np.shape(arr)[1]))  # scales/shrinks the data. Does not affect coordinate position of cells, utlimately not affecting final prediction
    # -------------------------------------- 
    
    # D_r here is a vector (in paper is a diagonal matrix) of sums of each row, size n cells 
    D_r = np.sum(FM, axis=1)
    
    # D_c here is a vector of sums of each column, size K (2P genes)
    D_c = np.sum(FM, axis=0) # D_c -> vector containing sum of each row in FM
        
    # order of multiplying following matrices DOES matter:
    
    # (1 / D_r ^1/2) * FM
    # update D_r to be D_r ^ -1/2
    D_r = np.sqrt(np.reciprocal(D_r))
    # nxn diagonal matrix * nxk matrix = nxk , essentially just multiply the the nth row in FM by nth value in D_r
    for n in range(len(D_r)):
        FM[n] *= D_r[n] 
        
    # ((1 / D_r ^1/2) * FM)  * (1 / D_c ^1/2)
    # update D_c to be D_c ^ -1/2
    D_c = np.sqrt(np.reciprocal(D_c))
    # nxk matrix * kxk diagonal matrix = nxk , essentially just multiply the the ith column in FM by ith value in D_c
    for i in range(len(D_c)):
        FM[:,i] *= D_c[i]
    # -------------------------------- End of Generating matrix of standardized relative frequencies
    print("Matrix of standardized relative frequencies:\n", FM, "\n", np.shape(FM))
    print("D_r ^-1/2:\n", D_r, "\n", np.shape(D_r))
    print("D_c ^-1/2:\n", D_c, "\n", np.shape(D_c))
    print("Fuzzy-coded indicator matrix:\n", X, "\n", np.shape(X))

    return {"Z": FM,    # S_N,K, matrix of standardized relative frequencies
            "D_r": D_r, # D_r ^-1/2
            "D_c": D_c, # D_c ^-1/2
            "X": X      # Fuzzy-coded indicator matrix
            }


def MCAStep2(S: np.ndarray, U: np.ndarray, D_r: np.ndarray, D_c: np.ndarray) -> dict:
    """
    Computes cell (row standard) and gene (column principal) coodinates
    input:
        - S: matrix of standardized relative frequencies
        - U: numpy 2D array of left singuar vectors as columns from SVD computation of S
        - D_r: numpy vector representing D_r ^-1/2
        - D_c: numpy vector representing D_c ^-1/2
    returns a dictionary:
        - cellCoordinates: numpy 2D array of cells coordinates where rows are cells and there are j number of columns where columns represent dimensions
        - geneCoordinates: numpy 2D array of gene coordinates where rows are features (genes) and there are j number of columns where columns represent dimensions
    """    
    
    ################## Row standard coordinates : Φ = D_r ^−1/2 * U = D_r ^−1/2 * S * V * D_a ^−1 ################    
    # Φ = D_r ^-1/2 * U
        # D_r size: n cells vector
        # U size: n cells x j dimensions
        # Φ size: n cells x j dimensions
    row_coordinates = np.empty(shape=(np.shape(U)))
    # https://solitaryroad.com/c108.html
    # diagonal matrix = [[k1,0,0], [0,k2,0], [0,0,k3]]
    # matrix = [[x11,x12,x13], [x21,x22,x23], [x31,x32,x33]]
    # diagonal matrix * matrix = [[k1*x11, k1*x12, k1*x13], [k2*x21, k2*x22, k2*x23], [k3*x31, k3*x32, k3*x33]]
    for i in range(np.shape(U)[1]):
        row_coordinates[:,i] = D_r * U[:,i]
    
    ################## Column principal coordinates : G = D_c ^−1/2 * S^T * U = D_c ^−1/2 * V * D_a #############
    # G = D_c ^−1/2 * S^Transposed * U
        # D_c size: k vector where k = 2P genes
        # S^T size: k x n cells where k = 2p genes
        # U   size: n cells x j dimensions
        # G   size: k x j dimensions where k = 2p genes
    S = np.transpose(S) # S^T
    column_coordinates = np.empty(shape=(np.shape(S)))
    # diagonal matrix * matrix = [[k1*x11, k1*x12, k1*x13], [k2*x21, k2*x22, k2*x23], [k3*x31, k3*x32, k3*x33]]
    for i in range(np.shape(column_coordinates)[1]): # (D_c ^−1/2 * S^Transposed)
        column_coordinates[:,i] = D_c * S[:,i]
    column_coordinates = np.dot(column_coordinates, U) # (D_c ^−1/2 * S^Transposed) * U
    
    # only keep + complememnt version of each gene
    rows, columns = np.shape(column_coordinates)
    ret = np.empty(shape=(int(rows/2), columns)) # fuzzy matrix ends up with total of K = 2P categories/columns
    index = 0
    for row in range(rows):
        if row % 2 == 0:
            ret[index] = column_coordinates[row]
            index += 1
    
    return {"cellCoordinates": row_coordinates,
            "geneCoordinates": ret }


import numpy as np
import pandas as pd
"""
    Copyright (C) 2023  Ramin Mohammadi

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses>."""
        
    
def filter_genes(df: pd.DataFrame, n: int):
    """
    Genes expressing at least one count in less than n cells are removed.

    input:
        - df: raw gene count data frame with genes as rows and cells as columns
        - n: min number of cells that the gene has to be expressed in
    output:
        - list of genes kept
    """    
    print(f"Removing genes that are expressed in less than {n} cells...")
    genes = []
    df = df.map(lambda x: 1 if x > 0 else 0) # turn all gene count values > 0 to 1
    for gene in df.index:
        if np.sum(df.loc[gene]) >= n: genes.append(gene) # if sum of a row >= n (more than n cells express that gene)
    return genes


import numpy as np
import pandas as pd
import csv, time, scipy
from Ramin123455 import mca_steps

"""
    Copyright (C) 2023  Ramin Mohammadi

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses>."""
    
#Reference: https://github.com/RausellLab/CelliD
    
    
# MCA CLASS responsible for holding the results of MCA, mainly being the cell and gene coordinates, 
# which are pandas dataframes each containing a 2D matrix of numerical values, and the fuzzy-coded indicator matrix
class MCA:
    def __init__(self, cellCoordinates: np.ndarray, geneCoordinates: np.ndarray, X: np.ndarray, 
                genesN: list, cellsN: list, j: int):
        mca_strings = [f"MCA_{i}" for i in range(1, j+1)]
        df1 = pd.DataFrame(cellCoordinates, index=cellsN, columns=mca_strings)
        df2 = pd.DataFrame(geneCoordinates, index=genesN, columns=mca_strings)        
        self.cellCoordinates = df1
        self.geneCoordinates = df2
        self.X = X # fuzzy-coded indicator matrix


def RunMCA(arr: pd.DataFrame, j: int = 50, genes: list = None, write_results_to_csv = False):
    """
    Performs MCA (multiple correspondence analysis) on input data (preprocessed raw gene count data), generating cell and gene coordinates in j dimensional space

    input:
        - arr: a pandas data frame with cells as rows and genes as columns
        - j: number of dimensions in coordinate space to compute and store, default set to 50
        - genes: String vector of gene names that want to be used. If not specified all genes will be used.
        - write_results_to_csv: If True, will write several results from MCA method to csv files.
    output:
        - MCA object containing:
            - cellCoordinates: pandas data frame where rows are cells and columns are j (default value for j is 50) different dimensions
            - geneCoordinates: pandas data frame where rows are genes and columns are j (default value for j is 50) different dimensions
            - X: fuzzy-coded indicator matrix
    """
    
    # preprocessing matrix
    # ------------------------------------------------
    
    # genes if passed a value should be a list of specific column names genes as strings that want to be used
    if (genes is not None):
        arr = arr.loc[:, genes]
    
    # Remove columns with all zeros
    arr = arr.loc[:, (arr != 0).any(axis=0)]

    # Remove columns with duplicated column names
    arr = arr.loc[:, ~arr.columns.duplicated()]
    
    # Get row names (cell names) as 'cellsN' and column names (gene names) as 'genesN' after the preprocessing step
    cellsN = arr.index
    genesN = arr.columns
    
    arr = arr.to_numpy()
    # -------------------------------------------------
    
    ######### Fuzzy Matrix computation- MCAStep1 #########
    start_time = time.time()
    
    print("Computing Fuzzy Matrix...")
    MCAPrepRes = mca_steps.MCAStep1(arr)
    
    end_time = time.time()
    elapsed_time = end_time - start_time
    print(f"Elapsed time: {elapsed_time:.6f} seconds\n") # Timer used to determine how long it took to run MCAStep1
    
    if write_results_to_csv:
        with open("Results_csv/Z_matrix_stand_rel_freq.csv", 'w', newline='') as csvfile:
            csv_writer = csv.writer(csvfile)
            for row in MCAPrepRes["Z"]:
                csv_writer.writerow(row)
        with open("Results_csv/D_r_neg_one_half.csv", 'w', newline='') as csvfile:
            csv_writer = csv.writer(csvfile)
            csv_writer.writerow(MCAPrepRes["D_r"])
        with open("Results_csv/D_c_neg_one_half.csv", 'w', newline='') as csvfile:
            csv_writer = csv.writer(csvfile)
            csv_writer.writerow(MCAPrepRes["D_c"])
        with open("Results_csv/X_fuzzy_coded_matrix.csv", 'w', newline='') as csvfile:
            csv_writer = csv.writer(csvfile)
            for row in MCAPrepRes["X"]:
                csv_writer.writerow(row)
    #######################################################
    
    ########## Singular Value Decomposition (SVD) computation ############      
    start_time = time.time()

    print("Computing SVD...")
    # partial Singular Value Decomposition        
    # input matrix size: n cells x d columns   (d = 2p genes)
    # parameter k: number of singular values and vectors to compute (default is 50)
    # SVD returns:
        # U, size: n x k, (left singular vectors as columns) 
        # S, size: k vector, (singular values)
        # V_Transposed, size: , (right singular vectors as rows)
    U, S, Vh = scipy.sparse.linalg.svds(MCAPrepRes["Z"], k=j, which='LM') 
    
    end_time = time.time()
    elapsed_time = end_time - start_time
    print(f"Elapsed time: {elapsed_time:.6f} seconds\n")
        
    # Note: In the context of Singular Value Decomposition (SVD), the sign of the values in the matrices does not matter for the mathematical correctness or validity of the decomposition. The singular value decomposition is unique up to a sign convention.
        
    # reshape and transpose Vh to allow MCAStep2() to multiply matrix of standardized relative frequencies by Vh
    # SVD returns Vh in transposed form so must undo to get right singular vectors as columns
    Vh = np.transpose(Vh)
    
    # put singular values in descending order 
    S = np.flip(S)
    
    print("SVD (Singular Value Decomposition) of standardized relative frequencies matrix:")
    print("U (left singular vectors as columns):\n", U, "\n", np.shape(U))
    print("S (singular values):\n", S, "\n", np.shape(S))
    print("Vh (right singular vectors as columns):\n", Vh, "\n", np.shape(Vh))
    
    if write_results_to_csv:
        with open("Results_csv/PartialSVD_U.csv", 'w', newline='') as csvfile:
            csv_writer = csv.writer(csvfile)
            for row in U:
                csv_writer.writerow(row)
        with open("Results_csv/PartialSVD_S.csv", 'w', newline='') as csvfile:
            csv_writer = csv.writer(csvfile)
            csv_writer.writerow(S)
        with open("Results_csv/PartialSVD_Vh.csv", 'w', newline='') as csvfile:
            csv_writer = csv.writer(csvfile)
            for row in Vh:
                csv_writer.writerow(row)
    #####################################################
    
    ############## Coordinate Computations- MCAStep2 ###############
    start_time = time.time()

    print(f"Computing Coordinates for a {j} dimensional space...")
    coordinates = mca_steps.MCAStep2(S = MCAPrepRes["Z"], U=U, D_r=MCAPrepRes["D_r"], D_c=MCAPrepRes["D_c"]) # main logic
    
    end_time = time.time()
    elapsed_time = end_time - start_time
    print(f"Elapsed time: {elapsed_time:.6f} seconds\n")
    
    # MCA is an object that stores gene and cell coordinates, and fuzzy coded indicator matrix
    mca = MCA(cellCoordinates=coordinates["cellCoordinates"], geneCoordinates=coordinates["geneCoordinates"],
                X=MCAPrepRes["X"], genesN=genesN, cellsN=cellsN, j=j)
    
    print("Cell coordinates:\n", mca.cellCoordinates)
    print("Gene coordinates:\n", mca.geneCoordinates)
    
    if write_results_to_csv:
        mca.geneCoordinates.to_csv('Results_csv/geneCoordinates.csv')
        mca.cellCoordinates.to_csv('Results_csv/cellCoordinates.csv')
    ################################################################
    
    return mca


import pandas as pd
import csv
import anndata as ad
import scanpy as sc

"""
    Copyright (C) 2023  Ramin Mohammadi

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses>."""

# scanpy gene ranking (DE)
# https://scanpy.readthedocs.io/en/stable/generated/scanpy.tl.rank_genes_groups.html


def top_diff_exp_genes(gene_count_df: pd.DataFrame, meta_data_df: pd.DataFrame, meta_list: list, 
                        meta_col_name: str, score_threshold=0, show_top_genes_plot=False, write_top_genes_to_csv=False):
    """
    Returns the most differentially expressed genes per meta data grouping of the cells using scanpy's rank_genes_groups() function.
    All of the genes are given a score per cell group. Only the genes that are above the score_threshold are kept (default is 0).
    Meta data per cell is required for the rank_genes_groups() function to base a varying condition off of among the gene counts
    MUST pass log transformed version of the RAW gene count data frame in gene_count_df

    input:
        - gene_count_df: log transformed version of the RAW gene count data where cells are rows and genes are columns
        - meta_data_df: data frame containing meta data per cell (cells are rows and columns are different meta data)
        - meta_list: the list of all possible values in the meta data column used
        - meta_col_name: the column name of the meta data being used in the meta data frame
        - score_threshold: genes above this value when ranked will be kept
        - show_top_genes_plot: scanpy plot of the ranked genes per group
        - write_top_genes_to_csv: if True, will write the the genes kept to a .csv file
    output:
        - list of genes (gene names) that were kept
    """    
    
    print("Finding most differentially expressed genes...\n")
    # Ann data, .obs should hold the meta data and .var should hold gene names,
    # .X holds log transformed raw gene count data
    adata = ad.AnnData(X=gene_count_df) 
    adata.obs_names = gene_count_df.index.to_list()
    adata.var_names = gene_count_df.columns.to_list()
    adata.obs[meta_col_name] = pd.Categorical(meta_data_df[meta_col_name])
    sc.tl.rank_genes_groups(adata, groupby=meta_col_name) # group by is required argument
        
    if show_top_genes_plot:
        sc.pl.rank_genes_groups(adata)
            
    # genes are ordered/ranked per group of cells by their given score so get top x genes per cell group
    most_diff_exp_genes_set = []
    for cell_group in meta_list:
        dedf = sc.get.rank_genes_groups_df(adata, group=cell_group)
        print("Group: ", cell_group, "\n", dedf)
        for index in dedf.index:
            # only keep genes that were given a score above the threshold to avoid including genes that were not not highly expressed in a group
            if dedf['scores'][index] > score_threshold: 
                most_diff_exp_genes_set.append(dedf['names'][index])

    # Get rid of duplicate genes listed by acquiring the SET of the list of genes
    most_diff_exp_genes_set = list(set(most_diff_exp_genes_set))
 
    if write_top_genes_to_csv:
        with open('most_diff_exp_genes.csv', "w", newline="") as file:
            writer = csv.writer(file)
            writer.writerow(most_diff_exp_genes_set)
            
    # most differentially expressed genes in different groups may overlap (present in more than one group)
    return most_diff_exp_genes_set


import numpy as np
import pandas as pd
from sklearn.metrics.pairwise import euclidean_distances

"""
    Copyright (C) 2023  Ramin Mohammadi

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses>."""

# Reference: https://github.com/RausellLab/CelliD


def GetDistances(cellCoordinates: pd.DataFrame, geneCoordinates: pd.DataFrame, X: np.ndarray = None, barycentric=False, 
                 cells_filter: list = None, genes_filter: list = None, write_dist_to_csv = False):
    """
    Euclidean distance calculation between gene and cell coordinates among j dimensions. 
    The closer a gene g is to a cell c, the more specific to such a cell it can be considered.

    input:
        - cellCoordinates
        - geneCoordinates 
        - genes_filter: String vector of genes names to subset from gene coordinates. If not specified will use all genes 
        - cells_filter: String vector of cell names to subset from cell coordinates. If not specified will use all cells 
        - write_dist_to_csv: if True, writes the resulting euclidean distance matrix to a .csv file
    return:
        - CellGeneDistances: numpy 2d array with genes as rows and cells as columns. Each value represents the distance between gene x and cell y
    """    
    
    # filter genes by a list of specific genes as strings that want to be used
    if (genes_filter is not None):
        geneCoordinates = geneCoordinates.loc[genes_filter]
    # filter cells by a list of specific cells as strings that want to be used
    if (cells_filter is not None):
        cellCoordinates = cellCoordinates.loc[cells_filter]
            
    print("\nCalculating distances between cells and genes...\n")
    # calculate euclidean distance between cells and genes
    CellGeneDistances = dist(cells=cellCoordinates, genes=geneCoordinates, X=X, barycentric=barycentric)
    print("Distance matrix:\n", CellGeneDistances)
    if write_dist_to_csv:
        CellGeneDistances.to_csv('Results_csv/CellGeneDistances.csv')
    
    return CellGeneDistances


# At this stage, only gene category coordinates g_p+ , conveying presence of gene expression relative to the maximum per gene, are retained for downstream analysis.
def dist(cells: pd.DataFrame, genes: pd.DataFrame, X: np.ndarray, barycentric: bool) -> np.ndarray:
    """
    Euclidean distance calculations between every gene and cell based off the gene and cell coordinates among j dimensions
    input: 
        - genes: pandas data frame for GENE coordinates
        - cells: pandas data frame for CELL coordinates 
        - X: fuzzy coded indicator matrix 
        - barycentric: if True, will calculate gene coordinates through barycentric relationship with cell coordinates using X
    returns: 
        -a 2D numpy array of distances between each pair of cells and genes where rows are genes (features) and columns are cells
    """
    
    if(not barycentric):
        # euclidian distance between gene p and cell n among all j dimensions
        # cell matrix: n cells x j dimensions
        # gene matrix: p genes x j dimensions
        euclidean_dist = euclidean_distances(genes.to_numpy(), cells.to_numpy())
 
    # barycentric relationship gene coordinates: g_kj = (1 / N∑n=1 x_nk) * N∑n=1 (x_nk ϕ_nj)
    else:       
        print("Finding gene coordinates through barycentric relationship: ")

        g = np.empty(shape=(np.shape(X)[1], np.shape(cells)[1])) 
        rows, cols = np.shape(g)
        x = np.sum(X, axis=0)
        for k in range(rows):
            for j in range(cols):
                g[k,j] = (1 / x[k] ) * np.sum(X[:,k] * cells.iloc[:, j])
        
        # take only g+ so even rows in g
        g_plus = np.empty(shape=(int(np.shape(X)[1] / 2), np.shape(cells)[1]))
        index = 0
        for row in range(0, np.shape(g)[0]):
            if row % 2 == 0:
                g_plus[index] = g[row]
                index += 1
        euclidean_dist = euclidean_distances(g_plus, cells.to_numpy()) 
                
    return pd.DataFrame(euclidean_dist, index = genes.index, columns = cells.index)


import numpy as np
import pandas as pd
import sys
from scipy.stats import hypergeom, false_discovery_control

"""
    Copyright (C) 2023  Ramin Mohammadi

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, see <https://www.gnu.org/licenses>."""
    
# Reference: https://github.com/RausellLab/CelliD


def RunCellHGT(DT: pd.DataFrame, gene_sets: pd.Series, n_genes = 200, minSize = 10, p_adjust = True, log_trans = False):    
    """
    Predicts cell type/gene set for each unknown cell using hypergeometric testing. First acquires the gene signature for every cell n 
    by keeping the n_genes smallest euclidean distance values per cell in the distance data frame DT. Then finds W_1,W_2, ... W_omega. 
    These are the genes in each gene set W_i that are in the genes list P, the genes in the gene expression matrix after any gene filtering 
    steps (W_i is a subset of P). The gene sets W_i that have < minSize number of genes are ignored. w is acquired being the intersection of 
    the genes between each cell's gene signature and W_i. Then, hypergeometric distribution is performed with a probability mass function 
    using variables described on line 120. Probability values (p-values) between each cell and gene set is acquired. Next, optionally adjust 
    p-values with Benjamini Hochberg correction on every cell among the gene sets and/or -log base 10 transform p-values (log transform only 
    scales the values and has no affect on the prediction results). Finally, cell's gene sets are predictied using rules described on line 155.
    input: 
        - DT: pandas data frame, with genes as rows and cells as columns, of the euclidean distances between the coordinates of genes and cells
        - gene_sets: pandas series where each index is labelled with the gene set/cell type name and each value is a list of genes in that gene set/cell type 
        - n_genes: number of closest genes to keep per cell in the euclidean distance data frame, DT, to represent a cell's gene signature
        - minSize: min number of genes that a gene set must have to keep it after taking subset of P
        - p_adjust: if TRUE, apply Benjamini Hochberg correction, adjusting the p-values from hypergeo probability mass function
        - log_trans: if TRUE, apply -log base 10 to p-values after BH correction (not necessary, just upscales small p-values to larger values)
    returns a pandas Series:
        - index are the names of the unknown cells and values are the cell type/gene set predictions per cell      
    """    
    print("\nRanking genes...")
    genes = DT.index
    cells = DT.columns
    
    # Obtain gene signature for each cell by keeping the n_genes closest genes to each cell using the distance matrix DT--------------------------
    # i is a matrix containing the row INDICES of n_genes genes with the smallest distance from each cell (column), sorted least to greatest for each column 
    # -> (GENE RANKINGS FOR EACH CELL by row INDICE)
    i = np.argsort(DT, axis=0)[range(n_genes), :] 
    # gene_signatures contains values 0 and 1 representing if the gene is within the cell's gene signature (top n_genes closest genes to that cell).
    # Each column represents a cell's gene signature with values of 1 to corresponding genes.
    gene_signatures = np.zeros(shape=(len(genes), len(cells))) # create matrix of all zeros
    for column_num in range(i.shape[1]):
        for row_num in range(i.shape[0]):
            gene_signatures[ i[row_num, column_num], column_num ] = 1
    gene_signatures = pd.DataFrame(gene_signatures, index=genes, columns=cells)
    # ------------------------------------------------------------------------
    
    # Gene marker dataset -----------------------------------------------------------------
    # Determine W_1,W_2, ... W_omega. These are the genes in each gene set W_i that are in the genes list P, the genes in the 
    # gene expression matrix after any gene filtering steps (W_i is a subset of P)
    nPathInit = len(gene_sets) # number of cell types/gene sets in gene marker dataset
    cell_type_list = gene_sets.index
    # Row by row (every cell type), filter the genes in the gene marker dataset to be only the genes contained in P
    for i in range(len(gene_sets)): # iterate cell type by cell type
        new_gene_list= []
        for gene in gene_sets.iloc[i]: # iterate gene by gene for ith cell type
            if gene in genes: # check if gene is a gene in the list of unknown cells
                new_gene_list.append(gene)
        gene_sets.iloc[i] = new_gene_list
        
    # GENE MARKER FILTERING STEP: Only keep the cell types/gene sets, where the number of genes 
    # in the cell type W_i after the above step is >= minSize (default is 10)
    # gene_sets now stores data in a dictionary: {"cell type name 1": [list of genes], "cell type name 2": [list of genes], ...}
    dict_gene_sets = {}
    for i in range(len(gene_sets)):
        if len(gene_sets.iloc[i]) >= minSize:
            dict_gene_sets.update({cell_type_list[i] : gene_sets.iloc[i]}) 
    gene_sets = dict_gene_sets
    nPathEnd = len(gene_sets)
    nFiltPath = nPathInit - nPathEnd
    # Stop Hypergeometric testing if there are no cell types in the gene reference dataset that have more than minSize (default val is 10) 
    # genes that are present in the input gene expression data
    if nPathEnd == 0:
        sys.exit(f"All gene sets (possible cell types using the user-chosen gene marker dataset) have less than {minSize} genes in common with the data")
    print(f"{nPathEnd} gene sets kept for hypergeometric testing out of {nPathInit} of the gene sets. {nFiltPath} gene sets were filtered as they had less than {minSize} genes present in the data")
    
    # Place gene sets (W) in columns (cell types as columns and genes in P as rows. Values of 1 represent the genes that are in the cell type/gene set) 
    # gene_sets will contain row INDICES of where the genes in each gene set (W) are located in the distance matrix's (DT) gene list P (since W is a subset of P)
    # W_len contains dictionary with length of gene list for each cell type (practically gene sets W in number format used for the probability mass function)
    W_len = {}         
    genes_list_p = genes.to_list() 
    for key in gene_sets:        
        gene_indice = []
        for gene in gene_sets.get(key):
            gene_indice.append(genes_list_p.index(gene)) # find index of gene in gene list P
        W_len.update({key : len(gene_indice)})
        gene_sets[key] = gene_indice 
    # IMPORTANT: W_binary_matrix contains 0s and 1s representing the genes in each gene set/cell type, W, that are in gene list P.
    # Value is 1 if the gene is present for that cell type. Rows are genes in gene list P, columns are cell types/gene sets
    W_binary_matrix = np.zeros(shape=(len(genes), len(W_len)))
    j = 0
    for key in gene_sets:
        for index in gene_sets.get(key):
            W_binary_matrix[index, j] = 1
        j += 1    
    W_binary_matrix =  pd.DataFrame(W_binary_matrix, index=genes, columns=list(W_len.keys()))
    # ------------------------------------------------------------------------
    
    # gene_signatures contains 0s and 1s representing which of the genes are the top n_genes closest to cell Y (represented by 1)
    # W_binary_matrix contains 0s and 1s representing which of the genes are present (represented by 1) in the cell type (cell types determined from gene marker dataset used)
     
    # Hypergeo ----------------------------------------------------------------
    # Perform probabiliy mass function, optionally benjamini hochberg correction and log transform, 
    # then predict cell type, and assign not confident predictions with "unassigned".
    
    # scipy hypergeom probability mass function: p(k, M, n, N) 
    # n = W_i (number of genes in gene set/cell type after subsetting it from P)
    # k = w_i,n (number of genes in intersection of W_i and cell n's gene signature)
    # M = P (number of genes in input data after any gene filtering steps done before hyper geo testing)
    # N = n_genes (number of genes in cell's gene signature)
    
    n = list(W_len.values())

    # k is a pandas data frame where rows are the unknown cells and columns are the cell types of the gene sets. 
    # IMPORTANT: Each value in k is a w, the number of genes overlapping between a cell's gene signature (top n_genes closest genes to the cell) and a gene set W
    # (the number of genes in each intersection of (cell_n's gene signature and W_i gene set)
    k = pd.DataFrame(np.dot(np.transpose(gene_signatures), W_binary_matrix), index=gene_signatures.columns, columns=W_binary_matrix.columns)
    k = k.astype("int")

    M = len(genes)
    
    N = n_genes

    # Hypergeometric distribution stored in A: probability values (between 0 and 1) that unknown cell x is cell type y 
    # rows are genes in P, columns are cell types/gene sets
    A = pd.DataFrame()
    
    # Probability mass function:
    print("Performing hypergeometric test...\n")    
    i = 0
    cells = list(k.columns)
    # iterate the cell types/gene sets
    for i in range(0, len(cells)):
        prb = hypergeom.pmf(k=k.iloc[:, i], M=M, n=n[i], N=N) # generate probability for all unknown cells with the ith cell type/gene set
        A[cells[i]] = prb # prb is a vector
    A.index = k.index
    print("Probabilities:")
    print(A.T)

    """
    PREDICTING CELL TYPE: 
    - A cell is considered as ENRICHED in those gene sets for which the hypergeometric test p-value is < 1e-02 (0.01) -> 
    only if performing benjamini hochberg correction on probability values or performing neither BH correction nor log transformation,
    OR if performing log transformation on the probability values / after BH correction, -log10 corrected p-value > 2.
    - Simplified: enrichment ranges
        - case 1: p-value < 0.01 : (neither BH correction nor log transform) or (just BH correction on p-values)
        - case 2: p-value > 2 : (only log transform the probability values) or (BH correction then log transform p-values)
    - When a disjointed classification is required (for a cell, there are multiple values/gene sets that are in the enriched range), 
        - case 1: a cell will be assigned to the gene set with the LOWEST significant corrected p-value. 
        - case 2: a cell will be assigned to the gene set with the LARGEST significant corrected p-value. 
    - If no significant hits are found (cell is not enriched in any of the gene sets), a cell will remain unassigned.
    """

    # Benjamini Hochberg multiple testing correction:
    # adjust p-values to control the false discovery rate.
    if p_adjust:
        # perform benjamini correction by rows, each individual unknown cell and its p-values among the gene sets.
        A = pd.DataFrame(false_discovery_control(A, method='bh', axis=1), index=A.index, columns=A.columns)
        print("Benjamini Hochberg Correction:\n", A.T)         

    # CELL TYPE PREDICTING
    # prediction will be a series where the key is the unknown cell name and value is the cell type prediction
    if (p_adjust and not log_trans) or (not p_adjust and not log_trans): # just benjamini correction or neither
        prediction = A.idxmin(axis=1)  # take smallest p value for each unknown cell
        for cell in prediction.index:
            if np.min(A.loc[cell]) >= 0.01: # assign cell's whose most significant p value is not in enriched range as unassigned
                prediction[cell] = "unassigned"
    else: # BH correction then log transform, or no BH corr and log transform
        # -log base10 transform p-values in A
        A = -np.log10(A)
        print("-Log base10 transformation:\n", A.T)   
        prediction = A.idxmax(axis=1) # take largest p value for each unknown cell
        for cell in prediction.index:
            if np.max(A.loc[cell]) <= 2: # assign cell's whose most significant p value is not in enriched range as unassigned
                prediction[cell] = "unassigned"
    print("\nCell type/Gene set predictions:\n", prediction)

    return prediction


import numpy as np
import pandas as pd
from sklearn import preprocessing

"""
    Copyright (C) 2023  Ramin Mohammadi

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses>."""


def preprocess(assay_df: pd.DataFrame, order=1, log=True, normalize=True, scale=True, add1=False):
    """
    Preprocess raw gene count data through natural log transformation, normalization, and scaling. Input should be a 
    pandas data frame with cells as rows and genes as columns. Can optionally choose order of these preprocessing steps and which to be done.
    Adding 1 to the raw gene count data before hand can help prevent log(0) = undefined.
    # uses numpy and sklearn to perform natural log transformation, normalize, and scale the data
    # IMPORTANT: assay_df must be a pandas data frame where rows are cells (samples) and columns are genes (features)

        
    input:
        - assay_df: log transformed version of the RAW gene count data where cells are rows and genes are columns
        - order: value specifying 1 of 2 orders of performing preprocessing
        - log: if False, will not perform natural log transformation on data
        - normalize: if False, will not normalize the data
        - scale: if False, will not scale the data
        - add1: Possibly improve results by adding 1 to data if preprocessing in 2nd order
    output:
        - pandas data frame after raw gene count data has been preprocessed where cells are rows and genes are columns
    """    
    
    print("Beginning to preprocess data...\n")
    print("Before preprocessing:\n", assay_df)
    df_index = assay_df.index
    df_columns = assay_df.columns
    
    # default preprocess order: normalize, natural log transform, scale
    if order == 1: 
        assay_df = norm(assay_df, normalize)
        assay_df = logTransform(assay_df, log)
        assay_df = scaleData(assay_df, scale)
  
    # natural log transform, normalize, scale
    else:
        assay_df = logTransform(assay_df, log)
        if add1:
            assay_df += 1 # may improve results
        assay_df = norm(assay_df, normalize)
        assay_df = scaleData(assay_df, scale)
        
    assay_df = pd.DataFrame(assay_df, index=df_index, columns=df_columns)
    print("After preprocessing:\n", assay_df)
    return assay_df

# Normalize (represent values between 0 and 1). 
# Normalizes across axis 1 (by defualt), so across values in a single row (a row here should represent a cell so a cell's gene counts are normalized among all of the genes)    
# The default norm for normalize() is L2, also known as the Euclidean norm. The L2 norm formula is the square root of the sum of the squares of each value.
def norm(assay_df: pd.DataFrame, normalize: bool):
    if normalize:
        assay_df = preprocessing.normalize(assay_df)
        print("Normalized:\n", assay_df)
    return assay_df

# Natural log (base e) transformation. Helps reduce range of values
def logTransform(assay_df: pd.DataFrame, log: bool):
    if log:
        assay_df = np.log(assay_df)
        print("Natural log transformation:\n", assay_df)
    return assay_df

# Scaling Data (The purpose is to ensure that all features contribute equally to the model and to avoid the domination of features with larger values.)
def scaleData(assay_df: pd.DataFrame, scale: bool):
    if scale:
        scaler = preprocessing.StandardScaler().fit(assay_df)
        assay_df = scaler.transform(assay_df)
        print("Scaled:\n", assay_df)
    return assay_df